use crossbeam::{
    channel::{Receiver, Sender},
    select,
};
use serde::Serialize;
use statig::prelude::*;
use std::sync::{mpsc::Receiver as MpscReceiver, Mutex};
use tauri::{Manager, PhysicalPosition, PhysicalSize, WebviewWindow, Window};
use underlayer::{Bounds, UnderlayEvent};

use crate::storage::Storage;

impl From<State> for Event {
    fn from(value: State) -> Self {
        Self::State(value)
    }
}

impl From<&State> for Event {
    fn from(value: &State) -> Self {
        Self::State(value.clone())
    }
}

#[derive(Debug)]
pub enum Event {
    State(State),
    Bounds(Bounds),
    Return,
}

#[derive(Clone)]
pub struct Overlay {
    bounds: Bounds,
    window: WebviewWindow,
    previous: State,
}

pub struct OverlayData {
    // pub auto_hide: bool,
}

fn wrap_underlay_rx(rx: MpscReceiver<UnderlayEvent>) -> Receiver<UnderlayEvent> {
    let (cb_underlay_tx, cb_underlay_rx) = crossbeam::channel::unbounded();
    std::thread::spawn(move || {
        while let Ok(event) = rx.recv() {
            cb_underlay_tx.send(event).ok();
        }
    });
    cb_underlay_rx
}

impl Overlay {
    pub fn initialize(window: WebviewWindow, target_title: &str) -> Sender<Event> {
        let overlay: Overlay = Self {
            window: window.clone(),
            bounds: Bounds::default(),
            previous: State::hidden(),
        };

        window.manage(Mutex::new(OverlayData {}));

        let mut fsm = Overlay::uninitialized_state_machine(overlay).init();

        let underlay_rx = wrap_underlay_rx(underlayer::register(target_title));
        let (ee_tx, ee_rx) = crossbeam::channel::unbounded();
        let (window_event_tx, window_event_rx) = crossbeam::channel::unbounded();

        window.on_window_event(move |event| {
            if let tauri::WindowEvent::Focused(focus) = event {
                window_event_tx.send(*focus).ok();
            }
        });

        std::thread::spawn(move || loop {
            select! {
                recv(underlay_rx) -> msg => {
                    match msg {
                        Ok(event) => {
                                Self::handle_underlay_event(&mut fsm, event);
                        },
                        Err(_) => break,
                    }
                },
                recv(ee_rx) -> msg => {
                    match msg {
                        Ok(event) => {
                                Self::handle_event(&mut fsm, event);
                        },
                        Err(_) => break,
                    }
                },
                recv(window_event_rx) -> msg => {
                    match msg {
                        Ok(focus) => Self::handle_overlay_focused(&mut fsm, focus),
                        Err(_) => break,
                    }
                }
            }
        });

        ee_tx
    }

    fn handle_overlay_focused(fsm: &mut InitializedStateMachine<Overlay>, focus: bool) {
        match fsm.state() {
            State::Visible {} => {
                if focus {
                    fsm.window.set_ignore_cursor_events(true).ok();
                    underlayer::focus_target();
                }
            }
            _ => (),
        }
    }

    fn handle_underlay_event(fsm: &mut InitializedStateMachine<Overlay>, event: UnderlayEvent) {
        match event {
            UnderlayEvent::Attach {
                bounds,
                has_access: _,
                is_fullscreen: _,
            } => {
                fsm.handle(&Event::Bounds(bounds));
            }
            UnderlayEvent::MoveResize(bounds) => fsm.handle(&Event::Bounds(bounds)),
            UnderlayEvent::Detach => fsm.handle(&State::hidden().into()),
            UnderlayEvent::Focus => {
                fsm.window.emit("overlay_or_target_focus", true).ok();
                if let State::Hidden {} = fsm.state() {
                    fsm.handle(&State::Visible {}.into());
                }
            }
            UnderlayEvent::Blur => {
                if !fsm.window.is_focused().unwrap() {
                    fsm.window.emit("overlay_or_target_focus", false).ok();
                    if fsm
                        .window
                        .state::<Mutex<Storage>>()
                        .lock()
                        .is_ok_and(|s| s.config.hide_on_unfocus)
                    {
                        fsm.handle(&State::Hidden {}.into())
                    }
                }
            }
            UnderlayEvent::X11FullscreenEvent { is_fullscreen: _ } => {}
        }
    }

    fn handle_event(fsm: &mut InitializedStateMachine<Overlay>, event: Event) {
        fsm.handle(&event);
    }
}

#[state_machine(
    initial = "State::hidden()",
    state(derive(Debug, Clone, PartialEq, Eq, Serialize)),
    superstate(derive(Debug)),
    on_transition = "Self::on_transition"
)]
impl Overlay {
    #[action]
    fn on_enter_visible(&mut self) {
        self.window.set_always_on_top(true).ok();
        self.window.show().ok();
        self.window.set_resizable(true).ok();
        if self.bounds.width > 0 && self.bounds.height > 0 {
            self.window
                .set_position(PhysicalPosition::new(self.bounds.x, self.bounds.y))
                .ok();
            self.window
                .set_size(PhysicalSize::new(self.bounds.width, self.bounds.height))
                .ok();
        }
    }

    #[action]
    fn on_enter_interactable(&mut self) {
        self.window.set_ignore_cursor_events(false).ok();
        self.window.set_focus().ok();
    }

    #[action]
    fn on_enter_hidden(&mut self) {
        self.window.hide().ok();
        self.window.set_resizable(false).ok();
    }

    #[action]
    fn on_exit_interactable(&mut self) {
        self.window.set_ignore_cursor_events(true).ok();
    }

    #[action]
    fn on_explicitly_visible(&mut self) {
        underlayer::focus_target();
    }

    #[state(superstate = "super_visible", entry_action = "on_explicitly_visible")]
    fn visible(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Handled,
                State::Hidden {} => Transition(State::hidden()),
                State::Interactable {} => Transition(State::interactable()),
            },
            _ => Super,
        }
    }

    #[state(superstate = "common", entry_action = "on_enter_hidden")]
    fn hidden(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Transition(State::visible()),
                State::Hidden {} => Handled,
                State::Interactable {} => Transition(State::interactable()),
            },
            _ => Super,
        }
    }

    #[state(
        superstate = "super_visible",
        entry_action = "on_enter_interactable",
        exit_action = "on_exit_interactable"
    )]
    fn interactable(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Transition(State::visible()),
                State::Hidden {} => Transition(State::hidden()),
                State::Interactable {} => Handled,
            },
            _ => Super,
        }
    }

    #[superstate(superstate = "common", entry_action = "on_enter_visible")]
    #[allow(unused_variables)]
    fn super_visible(&mut self, event: &Event) -> Response<State> {
        Super
    }

    #[superstate]
    fn common(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::Bounds(bounds) => {
                self.bounds = *bounds;
                if bounds.width > 0 && bounds.height > 0 {
                    self.window
                        .set_position(PhysicalPosition::new(bounds.x, bounds.y))
                        .ok();
                    self.window
                        .set_size(PhysicalSize::new(bounds.width, bounds.height))
                        .ok();
                }
                Super
            }
            Event::Return => {
                log::trace!("Returning to: {:?}", self.previous);
                Transition(self.previous.clone())
            }
            _ => Super,
        }
    }

    fn on_transition(&mut self, src: &State, dest: &State) {
        if src != dest {
            self.previous = src.clone();
        }
        self.window.emit("OverlayStateChange", dest).ok();
    }
}
