import { Injectable } from '@angular/core';
import { EMPTY, Observable, ReplaySubject, Subject, from, map, mergeMap, switchMap, tap } from 'rxjs';
import { Plan, PlanInterface, PlanMetadata } from '../_models/plan';
import { MatDialog } from '@angular/material/dialog';
import { UrlDialog } from '../plan-display/url-dialog.component';
import { fetch } from '@tauri-apps/plugin-http';
import { invoke } from '@tauri-apps/api/core';

export class UrlError extends Error {
  status: number;

  constructor(status: number, message: string) {
    super(message);
    this.status = status;
  }
}

@Injectable({
  providedIn: 'root'
})
export class PlanService {

  private _currentPlanSubject: Subject<Plan> = new ReplaySubject<Plan>(1);
  private _basePlanSubject: Subject<Plan> = new ReplaySubject<Plan>(1);
  private _storedPlansSubject: Subject<PlanMetadata[]> = new ReplaySubject<PlanMetadata[]>(1);

  constructor(private dialog: MatDialog) {
    this.loadBasePlan();
    this.loadStoredPlans();
  }

  public getBasePlan(): Observable<Plan> {
    return this._basePlanSubject.asObservable();
  }

  public getCurrentPlan(): Observable<Plan> {
    return this._currentPlanSubject.asObservable();
  }

  public setCurrentPlan(plan: Plan) {
    this._currentPlanSubject.next(plan);
  }

  public getStoredPlans(): Observable<PlanMetadata[]> {
    return this._storedPlansSubject.asObservable();
  }

  public loadPlanFromPath(path: string, save_local: boolean = true): Observable<Plan> {
    return from(invoke<PlanInterface>('load_plan_at_path', { path, saveLocal: save_local })).pipe(map(plan => {
      return new Plan(plan)
    })).pipe(tap(() => {
      this.loadStoredPlans();
    }));
  }

  public loadFromUrl(url?: string, name?: string, save_local: boolean = false): Observable<Plan> {
    if (!url || !name) {
      const dialogRef = this.dialog.open(UrlDialog, {
        data: {
          url: url,
          name: name
        }
      });

      return dialogRef.afterClosed().pipe(switchMap(data => {
        if (data.url) {
          return this._loadFromUrl(data.url, data.name, save_local);
        }

        return new Observable<Plan>((s) => s.complete());
      }));

    } else {
      return this._loadFromUrl(url, name, save_local);
    }
  }

  public savePlanAtPath(path: string, plan: Plan) {
    plan.plan.forEach(elem => {
      if (!elem.notes) { elem.notes = "" }
    });

    return from(invoke('save_plan_at_path', { path, plan: plan.toInterface() }));
  }

  public checkForPlanUpdate(plan: Plan | PlanMetadata): Observable<Plan> {
    if (!plan.update_url) return EMPTY;

    return from(fetch(
      plan.update_url,
      {
        method: 'HEAD',
        connectTimeout: 10 * 1000
      })).pipe(switchMap(response => {
        this.validateResponse(response);

        if (response.headers.get('etag') === plan.latest_server_etag) {
          return EMPTY;
        }

        return this._loadFromUrl(plan.update_url!, plan.name!, false);
      }));
  }

  public savePlanAtStore(name: string, plan: Plan, allowOverwrite: boolean = false) {
    return from(invoke<string>('save_plan_at_store', { name, plan: plan.toInterface(), allowOverwrite })).pipe(tap(() => {
      this.loadStoredPlans();
    }));
  }

  private _loadFromUrl(url: string, name: string, save_local: boolean): Observable<Plan> {
    //Tauri fetch
    return from(fetch(
      url,
      {
        method: 'GET',
        connectTimeout: 10 * 1000
      }))
      .pipe(tap(response => { this.validateResponse(response); }))
      .pipe(
        mergeMap(response => from(response.json())
          .pipe(map(data => { 
            return {data, response};
           })))
      ).pipe(map(({data, response}) => {

        const plan = new Plan(data);
        const etag = response.headers.get('etag');
        if (etag) {
          plan.latest_server_etag = etag;
          console.log("got etag: ", etag);
        }

        return plan;
      })).pipe(tap(plan => {
        plan.update_url = url;
        plan.name = name;

        if (save_local) {
          this.savePlanAtStore(name, plan).subscribe();
        }

        this.loadStoredPlans();
      }));
  }

  private validateResponse(response: Response) {
    if (!response.ok) {
      throw new UrlError(response.status, "Error fetching plan from URL, status " + response.status + ", URL: " + response.url);
    }
  }

  private loadBasePlan() {
    from(invoke<PlanInterface>('base_plan')).subscribe(plan => {
      plan.plan.forEach(elem => { elem.edited = false; });
      plan.current = 0;
      plan.name = "Base Plan";
      this._basePlanSubject?.next(new Plan(plan));
    });
  }

  private loadStoredPlans() {
    from(invoke<PlanMetadata[]>('enumerate_stored_plans')).subscribe(plans => {
      this._storedPlansSubject.next(plans);
    })
  }
}
